/*
 * GOATTRACKER sound routines
 */

#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <mem.h>
#include <conio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "gsid.h"
#include "goattrk.h"

#define MINSPEED 11025
#define MAXSPEED 48000
#define MAXBUFFERS 50
#define FRAMERATE 50
#define SND_UPDATE_INTERVAL 20

/* Configurable from command line */
int buffers;

int buffernum = 0;
int maxfragment;
int outsamplesize;
int bfragment;
int initted = 0;
int playspeed;

char *outbuffer[MAXBUFFERS] = {NULL};
HWAVEOUT hwaveout = NULL;
WAVEHDR waveouthdr[MAXBUFFERS] = {NULL};
int outbufferprepared[MAXBUFFERS] = {0};
WAVEFORMATEX waveoutfmt;

static HANDLE snd_mixerthread = NULL;
static DWORD snd_mixerthreadid;
static int snd_mixerexit = 0;
static int writehandle = -1;

int sound_init(unsigned b, unsigned mr, unsigned writer);
void sound_uninit(void);
static void sound_update(void);
static DWORD WINAPI snd_mixer(LPVOID parameter);
static int openwaveout(void);
static void closewaveout(void);


int sound_init(unsigned b, unsigned mr, unsigned writer)
{
  int c;

  if (writer)
  {
    writehandle = open("sidaudio.raw", O_WRONLY | O_BINARY | O_CREAT, S_IREAD | S_IWRITE);
  }
  buffers = b;
  playspeed = mr;
  if (playspeed < MINSPEED) playspeed = MINSPEED;
  if (playspeed > MAXSPEED) playspeed = MAXSPEED;

  if (buffers < 2) buffers = 2;
  if (buffers > MAXBUFFERS) buffers = MAXBUFFERS;

  maxfragment = MAXSPEED / FRAMERATE;
  bfragment = playspeed / FRAMERATE;

  sid_init(playspeed);

  for (c = 0; c < buffers; c++)
  {
    outbuffer[c] = malloc(maxfragment * sizeof(short));
    if (!outbuffer[c])
    {
      return 0;
    }
    memset(outbuffer[c], 0, maxfragment);
  }

  if (!openwaveout())
  {
    return 0;
  }

  buffernum = 0;
  initted = 1;

  /* Create a separate thread for mixing */

  snd_mixerexit = 0;
  snd_mixerthread = CreateThread(NULL, 0, snd_mixer, NULL, 0, &snd_mixerthreadid);
  if (!snd_mixerthread)
  {
    closewaveout();
    return 0;
  }
  SetThreadPriority(snd_mixerthread, THREAD_PRIORITY_TIME_CRITICAL);

  return 1;
}

void sound_uninit(void)
{
  int c;

  if (writehandle != -1)
  {
    close(writehandle);
    writehandle = -1;
  }

  if (!initted) return;

  /* Signal exit to mixer thread and wait for it to finish */

	if (snd_mixerthread)
	{
		snd_mixerexit = 1;

    for (;;)
    {
      DWORD exitcode;

      if (GetExitCodeThread(snd_mixerthread, &exitcode))
      {
        if (exitcode != STILL_ACTIVE) break;
      }
    }
    CloseHandle(snd_mixerthread);
    snd_mixerthread = NULL;
	}

  closewaveout();

  for (c = 0; c < buffers; c++)
  {
    if (outbuffer[c])
    {
      free(outbuffer[c]);
      outbuffer[c] = NULL;
    }
  }

  initted = 0;
}

static DWORD WINAPI snd_mixer(LPVOID parameter)
{
  int timecounter = 0;
  int updates;
  DWORD time;
  DWORD prevtime;

  time = timeGetTime();

  while (!snd_mixerexit)
  {
    sound_update();

    /* Timing */
    updates = 0;
    while (!updates)
    {
      prevtime = time;
      time = timeGetTime();
      timecounter += (time - prevtime);
      updates = timecounter / SND_UPDATE_INTERVAL;
      timecounter -= updates * SND_UPDATE_INTERVAL;
      if (!updates) Sleep(SND_UPDATE_INTERVAL - timecounter);
    }
  }
  return 0;
}

void sound_update(void)
{
  int workcount = 0;
  int sendcount = 0;
  int firstbuffer;
  int samples;

  if (!initted) return;

  firstbuffer = buffernum;
  while ((waveouthdr[buffernum].dwFlags & WHDR_DONE) && (workcount < buffers))
  {
    if (outbufferprepared[buffernum])
    {
      waveOutUnprepareHeader(hwaveout, &waveouthdr[buffernum], sizeof(WAVEHDR));
      outbufferprepared[buffernum] = 0;
    }
    playroutine();
    sid_dumpregs();
    samples = sid_fillbuffer((short *)waveouthdr[buffernum].lpData, bfragment);
    if (writehandle != -1)
    {
      write(writehandle, waveouthdr[buffernum].lpData, samples * outsamplesize);
    }
    waveouthdr[buffernum].dwBufferLength = samples * outsamplesize;
    workcount++;
    buffernum++;
    buffernum %= buffers;
  }

  buffernum = firstbuffer;
  while (sendcount < workcount)
  {
    waveouthdr[buffernum].dwFlags = 0;
    waveOutPrepareHeader(hwaveout, &waveouthdr[buffernum], sizeof(WAVEHDR));
    outbufferprepared[buffernum] = 1;

    waveOutWrite(hwaveout, &waveouthdr[buffernum], sizeof(WAVEHDR));
    sendcount++;
    buffernum++;
    buffernum %= buffers;
  }
}

static int openwaveout(void)
{
  int c;

  closewaveout();

  outsamplesize = 2;
  waveoutfmt.wFormatTag = WAVE_FORMAT_PCM;
  waveoutfmt.nChannels = 1;
  waveoutfmt.wBitsPerSample = 16;
  waveoutfmt.nSamplesPerSec = playspeed;
  waveoutfmt.nAvgBytesPerSec = outsamplesize * playspeed;
  waveoutfmt.nBlockAlign = outsamplesize;
  if (waveOutOpen(&hwaveout, WAVE_MAPPER, &waveoutfmt, NULL, NULL, CALLBACK_NULL | WAVE_ALLOWSYNC) != MMSYSERR_NOERROR) return 0;

  /* Set up waveout buffers */
  for (c = 0; c < buffers; c++)
  {
    waveouthdr[c].lpData = outbuffer[c];
    waveouthdr[c].dwBufferLength = bfragment * outsamplesize;
    waveouthdr[c].dwFlags = WHDR_DONE;
    waveouthdr[c].dwLoops = 0;
  }
  return 1;
}

static void closewaveout(void)
{
  int c;

  if (hwaveout)
  {
    waveOutReset(hwaveout);
    for (c = 0; c < MAXBUFFERS; c++)
    {
      if (outbufferprepared[c])
      {
        waveOutUnprepareHeader(hwaveout, &waveouthdr[c], sizeof(WAVEHDR));
        outbufferprepared[c] = 0;
      }
    }
    waveOutClose(hwaveout);
    hwaveout = NULL;
  }
}

